/*
 *  Copyright (C) 2014 Benjamin W. (bitbatzen@gmail.com)
 *
 *  This file is part of Netvisix.
 *
 *  Netvisix is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Netvisix is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Netvisix.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "HostInfoPopup.h"
#include "NetView.h"
#include "ui_HostInfoPopup.h"
#include "Net/Host.h"
#include "Net/NetUtil.h"
#include "GUI/VisibleHost.h"
#include "GUI/StatisticPopup.h"

#include <QtGui>
#include <QStyle>
#include <QStyleOption>


namespace Netvisix {

    HostInfoPopup::HostInfoPopup(VisibleHost* vHost, QWidget* parent) :
            QWidget(parent),
            ui(new Ui::HostInfoPopup) {

        this->vHost = vHost;

        ui->setupUi(this);

        setAttribute(Qt::WA_DeleteOnClose);

        setStyleSheet("QWidget#HostInfoPopup {border: 2px solid rgb(110, 110, 110); background-color: rgb(150, 150, 150);}");

        ui->buttonStatistic->setStyleSheet("background-color: rgb(200, 200, 200)");

        QPoint vHostPos = vHost->getPosition().toPoint();

        const std::vector<Tins::IPv4Address> addrListIPv4 = vHost->getHost()->getAddrListIPv4();
        const std::vector<Tins::IPv6Address> addrListIPv6 = vHost->getHost()->getAddrListIPv6();
        int addrCount = addrListIPv4.size() + addrListIPv6.size();
        if (vHost->getHost()->addrHW != NetUtil::zeroAddrHW) {
            addrCount++;
        }

        int heightAddition = addrCount * ui->labelAddr->height();
        resize(width(), height() + heightAddition);

        ui->buttonStatistic->move(ui->buttonStatistic->pos().x(), ui->buttonStatistic->pos().y() + heightAddition);

        float offsetX = 0;
        float offsetY = -height();
        if (vHostPos.x() + width() > parentWidget()->width()) {
            offsetX = -width();
        }
        if (vHostPos.y() - height() < 0) {
            offsetY = 0;
        }

        QPoint pos = vHostPos + QPoint(offsetX, offsetY);
        move(pos);

        fixedFont = QFontDatabase::systemFont(QFontDatabase::FixedFont);
        fixedFont.setPointSize(11);
        fixedFontBold = QFontDatabase::systemFont(QFontDatabase::FixedFont);
        fixedFontBold.setPointSize(11);
        fixedFontBold.setBold(true);

        ui->labelHostname->setFont(fixedFontBold);
        ui->labelAddr->setFont(fixedFontBold);
        ui->labelAddrField->setFont(fixedFont);

        NetView* nv = (NetView*) parent;
        if (vHost->getHost()->hostname != "") {
            std::string hostname = "";
            if (vHost->getHost()->hostnameFromReverseDNSLookp) {
                hostname += "(rDNS) ";
            }
            hostname += vHost->getHost()->hostname;
            ui->labelHostname->setText(hostname.c_str());
        }
        else if (nv->getReverseDNSLookupEnabled() == false) {
            ui->labelHostname->setFont(fixedFont);
            ui->labelHostname->setText("(rDNS disabled!)");
        }

        // the addresses
        int addrLabelIndex = 0;
        if (vHost->getHost()->addrHW != NetUtil::zeroAddrHW) {
            addAddr("HW", vHost->getHost()->addrHW.to_string(), addrLabelIndex);
            addrLabelIndex++;
        }

        for (const Tins::IPv4Address& ip : addrListIPv4) {
            addAddr("IPv4", ip.to_string(), addrLabelIndex);
            addrLabelIndex++;
        }

        for (const Tins::IPv6Address& ip : addrListIPv6) {
            addAddr("IPv6", ip.to_string(), addrLabelIndex);
            addrLabelIndex++;
        }
    }

    HostInfoPopup::~HostInfoPopup() {
        delete ui;
    }

    void HostInfoPopup::addAddr(const std::string& type, const std::string& addr, int addrLabelIndex) {
        int posY = ui->labelAddr->pos().y() + addrLabelIndex * (ui->labelAddr->height());

        QLabel* labelType = new QLabel(type.c_str(), this);
        int typePosX = ui->labelAddr->pos().x();
        labelType->move(typePosX, posY);
        labelType->setFont(fixedFontBold);

        QLabel* labelAddr = new QLabel(addr.c_str(), this);
        int addrPosX = ui->labelAddrField->pos().x();
        labelAddr->move(addrPosX, posY);
        labelAddr->setFont(fixedFont);
    }

    void HostInfoPopup::paintEvent(QPaintEvent* event) {
        QPainter p(this);
        QStyleOption opt;
        opt.init(this);
        style()->drawPrimitive(QStyle::PE_Widget, &opt, &p, this);
    }

} // namespace Netvisix

void Netvisix::HostInfoPopup::on_buttonStatistic_clicked() {
    Host* host = vHost->getHost();
    std::string title = "";

    if (host->getAddrListIPv4().empty() == false) {
        title += host->getAddrListIPv4().at(0).to_string();
    }
    if (host->getAddrListIPv6().empty() == false) {
        if (title != "") {
            title += " | ";
        }
        title += host->getAddrListIPv6().at(0).to_string();
    }
    if (host->addrHW != NetUtil::zeroAddrHW) {
        if (title != "") {
            title += " ";
        }
        title += "(" + host->addrHW.to_string() + ")";
    }

    StatisticPopup* statisticPopup = new StatisticPopup(vHost->getHost()->statistic, title, parentWidget()->parentWidget()->parentWidget());
    statisticPopup->show();
}
